# coding: utf-8
"""
Модуль хэлперов приложения.
"""

import base64
import datetime
import os
import socket
import suds
import zlib
import logging
import traceback

from collections import namedtuple
from dateutil.relativedelta import relativedelta

import settings as gs_settings
from django.db.models import Q
from django.template import loader
from django.utils import simplejson as json

from suds.client import Client
from suds.sax.element import Element

from m3.actions.exceptions import ApplicationLogicException

from kinder.core.declaration.models import Declaration, DeclarationUnit
from kinder.core.queue_module.helpers import monthdelta
from kinder.core.dict.models import UnitKind

from kinder.core.queue_module import COMMON_QUEUE_STATUSES, DIRECT_STATUSES_TO_SHOW_IN_QUEUE
from kinder.core.direct.models import Direct

import queries as qs
import constants as const
from models import ReportFormRow, GisdoUnit
from gisdo.alchemy_queries import ApplicationIndex, EnrolledIndex
from gisdo.alchemy_session import session_scope

log = logging.getLogger('gisdo')


class SendReportException(Exception):
    """Ошибка отправки отчета"""


# --------------------------------------------
# Функции-хелперы
# --------------------------------------------
def get_count_direct(unit_id, date, ages_const):
    '''
    Задел на будущее.
    Чтобы считать  количество не по очереди, а напрямую по заявкам
    '''
    temporary_filter = Q(children__pupil__temporary=True)
    without_place_filter = ~(temporary_filter | Q(status__code=u'want_change_dou'))

    age1, age2 = const.AGE_CATEGORIES_CALC[ages_const]
    date1 = monthdelta(date, - int(12 * age2))
    date2 = monthdelta(date, - int(12 * age1))
    age_filter = Q(Q(children__date_of_birth__gt=date1) &
                   Q(children__date_of_birth__lte=date2))
    unit_and_date_filter = Q(
        #TODO: по мо другая фильтрация
        mo=unit_id,
        desired_date__lte=datetime.date(date.year, 12, 31))

    count = Declaration.objects.filter(without_place_filter & age_filter &
                                       unit_and_date_filter).count()
    return count


def deepupdate(original, update, merge=lambda x, y: x + y):
    """
    По умолчания складывает все значения словаря update и original
    @param original:
    @param update:
    @param merge:
    @return:
    """
    for key, value in original.iteritems():
        if not key in update:
            update[key] = value
        elif isinstance(value, dict):
            deepupdate(value, update[key])
        else:
            update[key] = merge(update[key], value)
    return update


# --------------------------------------------
# Классы-хелперы
# --------------------------------------------
row_cache = namedtuple('RowCache', ['applications',
                                    'queues',
                                    'enrolled',
                                    'capacities'])

#TODO: Антипаттерн God класс
class ReportFormHelper(object):
    """
    Класс-хелпер для работы с объектами ReportForm
    """

    MINNING_MODE = 1
    SEND_MODE = 2

    def __init__(self, report_form, schedule_settings, mode=MINNING_MODE):
        #: ссылка на отчет
        self.report_form = report_form

        #: шаблон xml
        self.template = 'ws/gisdo_request_xml_template.xml'
        self.file_name_pattern = u'last_request_xml_(%s).xml'

        #: ссылка на главное учреждение
        self.unit = report_form.unit

        #: дочерние учреждения
        self.descendants = self.unit.get_descendants(include_self=True)

        #: общее количество всех учреждений
        self.unit_count = self.descendants.filter(
            kind__id__in=[UnitKind.DOU, UnitKind.MO, UnitKind.REGION]).count()

        #: учреждения с типом UnitKind.MO
        self.mo_list = (self.descendants
                        .filter(kind__id__in=[UnitKind.MO, UnitKind.REGION])
                        .order_by('kind__id'))

        #: является ли главное учреждение регионом
        self.is_region = (self.unit.kind.id == UnitKind.REGION)

        self.cache = {}

        #: счетчики для мо, суммируют по всем ДОУ
        self.cache_mo = row_cache({}, {}, {}, {})

        #: кэш для подсчета данных по региону
        self.cache_region = row_cache({}, {}, {}, {}) if self.is_region else None

        #: настройки планировщика
        self.schedule_settings = schedule_settings

        # Кэш очереди
        if mode == self.MINNING_MODE:
            self.queue_cache = self._get_queue_cache()

    @staticmethod
    def get_list_dou(unit):
        """
        МО и его дочерние учреждения по порядку

        Запускается для Unit с типом UnitKind.MO или UnitKind.REGION
        """
        if unit.kind.id == UnitKind.MO:
            not_ready_dou_filter = Q(gisdo__not_on_federal_report=False) & Q(build_in_progress=False)
            result = (unit.get_descendants(include_self=True).filter(kind__id__in=[UnitKind.DOU, UnitKind.MO])
                                                             .filter(not_ready_dou_filter).order_by('kind__id'))
        else:
            result = [unit]
        return result

    def _get_queue_cache(self):
        """
        Очень тяжелый запрос выполняю один раз для всех учреждений и возрастных категорий
        Результат храним в памяти.
        В резальтате получаем словарь с разбивкой по учреждениям и возрастным категориям.
        """

        def _get_most_priority_unit(declaration):

            declaration_units = DeclarationUnit.objects.filter(declaration__id=declaration['id'],
                                                               unit__kind__id=UnitKind.DOU).order_by('ord')
            return declaration_units[0] if declaration_units.exists() else None
        #TODO: Брать статусы не из queue_module, declaration_status.models и direct.models
        declarations = Declaration.objects.filter(status__code__in=COMMON_QUEUE_STATUSES).select_related('children',
                                                                                                         'best_privilege',
                                                                                                         'work_type',
                                                                                                         'status')
        direct = Direct.objects.exclude(status__code__in=DIRECT_STATUSES_TO_SHOW_IN_QUEUE).values('declaration_id')
        if self.unit.kind.id != UnitKind.REGION:
            declarations = declarations.filter(mo=self.unit)
        declarations = declarations.filter(~Q(id__in=direct)).values('id', 'work_type__code', 'status__code',
                                                                     'desired_date', 'best_privilege__id',
                                                                     'children__health_need_id',
                                                                     'children__date_of_birth', 'mo')
        queue_cache = {}
        age_categories = const.AGE_CATEGORIES_FULL.keys()
        age_categories.extend(const.AGE_CATEGORIES_CUT.keys())
        age_categories.extend(const.AGE_CATEGORIES_FULL_CURRENT.keys())
        age_categories.append('ALL')
        declarations = list(declarations)
        for declaration in declarations:

            # TODO: Запрос в цикле. Если возникают проблемы с производительностью, то переносим память.
            most_priority_unit = _get_most_priority_unit(declaration)

            if most_priority_unit:
                if not most_priority_unit.unit.dou_type or most_priority_unit.unit.dou_type.code == const.GOVERMENT:
                    dou_type = const.GOVERMENT
                elif most_priority_unit.unit.dou_type.code == const.NOT_GOVERMENT:
                    dou_type = const.NOT_GOVERMENT
                elif most_priority_unit.unit.dou_type.code == const.IP:
                    dou_type = const.IP

                if most_priority_unit.unit.id not in queue_cache:
                    queue_cache[most_priority_unit.unit.id] = {}
                    queue_cache[most_priority_unit.unit.id][dou_type] = {}

                date_of_birth = declaration['children__date_of_birth']

                for age_category in age_categories:

                    if age_category not in queue_cache[most_priority_unit.unit.id][dou_type]:
                        queue_cache[most_priority_unit.unit.id][dou_type][age_category] = []

                    if age_category in const.AGE_CATEGORIES_FULL_CURRENT:
                        birth_date_calculate = datetime.date.today()
                    else:
                        birth_date_calculate = const.CALC_DATE()

                    down_border = birth_date_calculate - relativedelta(years=const.AGE_CATEGORIES_CALC[age_category][1],
                                                                    months=const.AGE_CATEGORIES_CALC[age_category][3])
                    up_border = birth_date_calculate - relativedelta(years=const.AGE_CATEGORIES_CALC[age_category][0],
                                                                     months=const.AGE_CATEGORIES_CALC[age_category][2])

                    if down_border < date_of_birth <= up_border:
                        queue_cache[most_priority_unit.unit.id][dou_type][age_category].append(declaration)

        return queue_cache

    def collect_applications_data(self, unit):
        applications = {
            'DOO_TYPE': {const.DOO_TYPES['NON_STATE']: {}, const.DOO_TYPES['BUSINESS']: {}},
            'DELIVERY': {},
            'GROUP_TYPE': {const.GROUP_TYPES['SHORT_STAY']: {}},
            'STATUS': {}
        }
        with session_scope() as session:
            index = ApplicationIndex(unit, session)
            for value in const.AGE_CATEGORIES_FULL.values():
                #: Общее количество поданных заявлений на постановку в очередь
                applications.setdefault('COMMON', {})[value] = index.get_count(age_range=value)
                #: Общее количество поданных заявлений на постановку в очередь в негосударственные ДОО
                applications['DOO_TYPE'][const.DOO_TYPES['NON_STATE']][value] = index.get_count(dou_type=
                                                                                                const.NOT_GOVERMENT,
                                                                                                age_range=value)
                #: Общее количество поданных заявлений на постановку в очередь у ИП
                applications['DOO_TYPE'][const.DOO_TYPES['BUSINESS']][value] = index.get_count(dou_type=
                                                                                               const.IP,
                                                                                               age_range=value)
            #: Общее количество заявлений поданных с Портал
            applications['DELIVERY'][const.DELIVERY_TYPES['PORTAL']] = (index
                                                                        .get_count(index_type=
                                                                                   ApplicationIndex.COUNT_BY_DELIVERY,
                                                                                   portal=1))
            #: Общее количество заявлений поданных при личном обращении
            applications['DELIVERY'][const.DELIVERY_TYPES['IN_PERSON']] = (index
                                                                           .get_count
                                                                           (index_type=
                                                                            ApplicationIndex.COUNT_BY_DELIVERY))
            #: Общее количество поданных заявлений на постановку в очередь в группы кратковременного пребывания детей
            for value in const.AGE_CATEGORIES_CUT.values():
                applications['GROUP_TYPE'][const.GROUP_TYPES['SHORT_STAY']][value] = (index
                                                                                      .get_count
                                                                                      (index_type=
                                                                                       ApplicationIndex
                                                                                       .COUNT_BY_GROUP_TYPE,
                                                                                       age_range=value))
            applications['STATUS'][const.APPLICATION_STATUSES['DENIED']] = index.get_count(index_type=
                                                                                           ApplicationIndex.COUNT_DENIED
                                                                                           )
        deepupdate(applications, self.cache_mo.applications)

        return applications

    def collect_queues_data(self, unit):
        queues = {
            'STATUS': {
                const.APPLICATION_STATUSES['IN_LINE_DEFERRED']: {},
                const.APPLICATION_STATUSES['IN_LINE_WANT_CHANGE_DOO']: {},
                
            },
            'HEALTH_NEEDS': {const.HEALTH_NEEDS['DISABLED']: {}},
            'BENEFITS': {const.BENEFIT_TYPES['HAVE_BENEFITS']: {}},
            'DOO_TYPE': {
                const.DOO_TYPES['BUSINESS']: {},
                const.DOO_TYPES['NON_STATE']: {},
                const.NO: {}},
            'GROUP_TYPE': {const.GROUP_TYPES['SHORT_STAY']: {}},
        }
        with session_scope() as session:
            index = EnrolledIndex(unit=unit, session=session)
            for key, value in const.AGE_CATEGORIES_FULL.iteritems():
                #: Количество детей в очереди (отложенный спрос)
                queues['STATUS'][const.APPLICATION_STATUSES['IN_LINE_DEFERRED']][value] = qs.postponed_queue_children_count(self, unit.id, key)
                #: Количество детей в очереди, необеспеченных местом в ДОО(актуальный спрос)
                queues['DOO_TYPE'][const.NO][value] = qs.children_without_place_count(self, unit.id, key)
                #: Количество детей в очереди (с учетом детей, желающих сменить ДОО)
                queues['STATUS'][const.APPLICATION_STATUSES['IN_LINE_WANT_CHANGE_DOO']][value] = (queues['DOO_TYPE'][const.NO][value] +
                    index.get_count(age_range=value, index_type=EnrolledIndex.COUNT_WANT_CHANGE_DOU))
                #: Количество детей в очереди, имеющих право на льготное зачисление
                queues['BENEFITS'][const.BENEFIT_TYPES['HAVE_BENEFITS']][value] = qs.count_of_children_in_privilege_queue(self, unit.id, key)
                #: Количество детей в очереди с ограниченными возможностями здоровья
                queues['HEALTH_NEEDS'][const.HEALTH_NEEDS['DISABLED']][value] = qs.children_with_health_need_count(self, unit.id, key)
                #: Количество детей в очереди в негосударственные ДОО
                queues['DOO_TYPE'][const.DOO_TYPES['NON_STATE']][value] = qs.children_without_place_count(self, unit.id,
                                                                                      key, dou_type=const.NOT_GOVERMENT)
                #: Количество детей в очереди у ИП
                queues['DOO_TYPE'][const.DOO_TYPES['BUSINESS']][value] = qs.children_without_place_count(self, unit.id,
                                                                                      key, dou_type=const.IP)
            for key, value in const.AGE_CATEGORIES_FULL_CURRENT.iteritems():
                #: Количество детей в очереди, необеспеченных местом в ДОО(актуальный спрос)
                queues['DOO_TYPE'][const.NO][value] = qs.children_without_place_count(self, unit.id, key)
                #: Количество детей в очереди (отложенный спрос)
                queues['STATUS'][const.APPLICATION_STATUSES['IN_LINE_DEFERRED']][value] = qs.postponed_queue_children_count(self, unit.id, key)
            #: Количество детей в очереди, имеющих право на льготное зачисление на основании федеральных льгот
            queues['BENEFITS'][const.BENEFIT_TYPES['FEDERAL']] = qs.count_of_children_in_privilege_queue(self, unit.id, 'ALL', qs.FEDERAL)
            #: Количество детей в очереди, имеющих право на льготное зачисление на основании региональных льгот
            queues['BENEFITS'][const.BENEFIT_TYPES['REGIONAL']] = qs.count_of_children_in_privilege_queue(self, unit.id, 'ALL', qs.REGIONAL)
            #: Количество детей в очереди, нуждающихся в компенсирующих группах
            queues['HEALTH_NEEDS'][const.HEALTH_NEEDS['CORRECT']] = qs.children_with_health_need_count(self, unit.id, 'ALL', comp=1)
            #: Количество детей в очереди, нуждающихся в оздоровительных группах
            queues['HEALTH_NEEDS'][const.HEALTH_NEEDS['HEALTH']] = qs.children_with_health_need_count(self, unit.id, 'ALL', comp=2)
            #: Количество детей в очереди в группы кратковременного пребывания
            for key, value in const.AGE_CATEGORIES_CUT.iteritems():
                queues['GROUP_TYPE'][const.GROUP_TYPES['SHORT_STAY']][value] = qs.count_of_children_in_short_stay(self, unit.id, key)

        if not gs_settings.NO_CONSIDER_PRIORITY:
            deepupdate(queues, self.cache_mo.queues)
        return queues

    def collect_enrolled_data(self, unit):
        enrolled = {
            'COMMON': {},
            'GROUP_TYPE': {
                const.GROUP_TYPES['DISABLED']: {},
                const.GROUP_TYPES['CARE']: {
                    const.DOO_TYPES['NON_STATE']: {},
                    const.DOO_TYPES['BUSINESS']: {}
                },
                const.GROUP_TYPES['SHORT_STAY']: {}
            },
            'DOO_TYPE': {
                const.DOO_TYPES['NON_STATE']: {},
                const.DOO_TYPES['BUSINESS']: {}
            },
            'STATUS': {const.APPLICATION_STATUSES['IN_LINE_ENROLLED']: {}},
            'ALL_CHILDS': {}
        }
        with session_scope() as session:
            index = EnrolledIndex(unit, session)
            for key, value in const.AGE_CATEGORIES_FULL.iteritems():
                #: Количество зачисленных детей
                enrolled['COMMON'][value] = index.get_count(age_range=value)
                #: Количество детей, зачисленных в группы для детей с ограниченными возможностями здоровья
                enrolled['GROUP_TYPE'][const.GROUP_TYPES['DISABLED']][value] = (index
                                                                                .get_count(age_range=value,
                                                                                           index_type=
                                                                                           EnrolledIndex.COUNT_DISABLED)
                                                                                )
                #: Количество детей, зачисленных  в негосударственные ДОО
                enrolled['DOO_TYPE'][const.DOO_TYPES['NON_STATE']][value] = index.get_count(age_range=value,
                                                                                            dou_type=
                                                                                            const.NOT_GOVERMENT)
                #: Количество детей, зачисленных у ИП, реализующих программы дошкольного
                enrolled['DOO_TYPE'][const.DOO_TYPES['BUSINESS']][value] = index.get_count(age_range=value,
                                                                                           dou_type=
                                                                                           const.IP)
                #: Количество детей
                enrolled['STATUS'][const.APPLICATION_STATUSES['IN_LINE_ENROLLED']][value] = (enrolled['COMMON'][value] +
                                                                                             qs.children_without_place_count(self, unit.id, key))
                #: Общее количество детей в системе - данный показатель не отправляется на шину
                enrolled['ALL_CHILDS'][value] = (enrolled['STATUS'][const.APPLICATION_STATUSES['IN_LINE_ENROLLED']]
                                                 [value] + qs.postponed_queue_children_count(self, unit.id, key))

            for key, value in const.AGE_CATEGORIES_FULL_CURRENT.iteritems():
                #: Количество зачисленных детей c расчетом возраста на текущую дату
                enrolled['COMMON'][value] = index.get_count(age_range=value)

            #: Количество детей, зачисленных в семейные дошкольные группы
            enrolled['GROUP_TYPE'][const.GROUP_TYPES['FAMILY']] = index.get_count(index_type=EnrolledIndex.COUNT_FAMILY)
            #: Количество детей, зачисленных в компенсирующих группах
            enrolled['GROUP_TYPE'][const.GROUP_TYPES['CORRECT']] = index.get_count(index_type=
                                                                                   EnrolledIndex.COUNT_CORRECT)
            #: Количество детей, зачисленных в оздоровительных группах
            enrolled['GROUP_TYPE'][const.GROUP_TYPES['HEALTH']] = index.get_count(index_type=
                                                                                  EnrolledIndex.COUNT_HEALTH)
            #: Количество детей, желающие сменить ДОО
            enrolled['STATUS'][const.APPLICATION_STATUSES['WANT_CHANGE_DOO']] = index.get_count(index_type=
                                                                                                EnrolledIndex.
                                                                                                COUNT_WANT_CHANGE_DOU)
            for key, value in const.AGE_CATEGORIES_CUT.iteritems():
                #: Количество детей, зачисленных в группы по присмотру и уходу в негосударственных ДОО
                enrolled['GROUP_TYPE'][const.GROUP_TYPES['CARE']][const.DOO_TYPES['NON_STATE']][const.AGE_CATEGORIES_CUT[key]] = \
                    index.get_count(age_range=value, dou_type=const.NOT_GOVERMENT)
                #: Количество детей, зачисленных в группы по присмотру и уходу в ИП
                enrolled['GROUP_TYPE'][const.GROUP_TYPES['CARE']][const.DOO_TYPES['BUSINESS']][const.AGE_CATEGORIES_CUT[key]] = \
                    index.get_count(age_range=value, dou_type=const.IP)
                #: Количество детей, зачисленных в группы кратковременного пребывания
                enrolled['GROUP_TYPE'][const.GROUP_TYPES['SHORT_STAY']][const.AGE_CATEGORIES_CUT[key]] = \
                    index.get_count(age_range=value, index_type=EnrolledIndex.COUNT_SHORT_STAY)
        # сохраним значения для мо
        deepupdate(enrolled, self.cache_mo.enrolled)
        return enrolled

    def collect_capacities_data(self, unit):
        capacities = {
            'COMMON': {},
            'GROUP_TYPE': {const.GROUP_TYPES['DISABLED']: {}, const.GROUP_TYPES['SHORT_STAY']: {}},
            'DOO_TYPE': {const.DOO_TYPES['NON_STATE']: {}, const.DOO_TYPES['BUSINESS']: {}}
        }
        index = qs.CapacitiesIndex(unit)
        for value in const.AGE_CATEGORIES_CUT.values():
            capacities['COMMON'][value] = index.get_count(age_category=value)
            capacities['GROUP_TYPE'][const.GROUP_TYPES['DISABLED']][value] = (index.get_count
                                                                              (index_type=
                                                                               qs.CapacitiesIndex.HEALTH_NEED,
                                                                               age_category=value))
            capacities['GROUP_TYPE'][const.GROUP_TYPES['SHORT_STAY']][value] = (index.get_count
                                                                                (index_type=
                                                                                 qs.CapacitiesIndex.SHORT_STAY,
                                                                                 age_category=value))
            capacities['DOO_TYPE'][const.DOO_TYPES['NON_STATE']][value] = index.get_count(dou_type=const.NOT_GOVERMENT,
                                                                                          age_category=value)
            capacities['DOO_TYPE'][const.DOO_TYPES['BUSINESS']][value] = index.get_count(dou_type=const.IP,
                                                                                         age_category=value)
        #сохраним значения для мо
        deepupdate(capacities, self.cache_mo.capacities)
        return capacities

    @staticmethod
    def get_applications_data(applications):
        """
        Данные по помещениям
        """
        result = []
        for category in const.AGE_CATEGORIES_FULL.values():
            #: Общее количество поданных заявлений на постановку в очередь
            result.append({
                'age_category': category,
                'health_needs': const.ALL,
                'benefit':      const.ALL,
                'status':       const.ALL,
                'delivery':     const.ALL,
                'group_type':   const.ALL,
                'doo_type':     const.ALL,
                'count':        applications['COMMON'][category]
            })
            #: Общее количество поданных заявлений на постановку в очередь в негосударственные ДОО (ИП)
            for doo_type in const.DOO_TYPES.values():
                result.append({
                    'age_category': category,
                    'health_needs': const.ALL,
                    'benefit':      const.ALL,
                    'status':       const.ALL,
                    'delivery':     const.ALL,
                    'group_type':   const.ALL,
                    'doo_type':     doo_type,
                    'count':        applications['DOO_TYPE'][doo_type][category]
                })
        for delivery in const.DELIVERY_TYPES.values():
            #: Общее количество заявлений поданных с Портала (При личном обращении)
            result.append({
                'age_category': const.ALL,
                'health_needs': const.ALL,
                'benefit':      const.ALL,
                'status':       const.ALL,
                'delivery':     delivery,
                'group_type':   const.ALL,
                'doo_type':     const.ALL,
                'count':        applications['DELIVERY'][delivery]
            })
        for category in const.AGE_CATEGORIES_CUT.values():
            #: Общее количество поданных заявлений на постановку в очередь в группы кратковременного пребывания детей
            result.append({
                'age_category': category,
                'health_needs': const.ALL,
                'benefit':      const.ALL,
                'status':       const.ALL,
                'delivery':     const.ALL,
                'group_type':   const.GROUP_TYPES['SHORT_STAY'],
                'doo_type':     const.ALL,
                'count':        applications['GROUP_TYPE'][const.GROUP_TYPES['SHORT_STAY']][category]
            })
        result.append({
            #: Количество отказов
            'age_category': const.ALL,
            'health_needs': const.ALL,
            'benefit':      const.ALL,
            'status':       const.APPLICATION_STATUSES['DENIED'],
            'delivery':     const.ALL,
            'group_type':   const.ALL,
            'doo_type':     const.ALL,
            'count':        applications['STATUS'][const.APPLICATION_STATUSES['DENIED']]
        })
        return result

    @staticmethod
    def get_queues_data(queues):
        """
        Данные по детям в очереди
        """
        result = []
        for category in const.AGE_CATEGORIES_FULL.values():
            result.extend([
                #: Количество детей в очереди (отложенный спрос)
                {
                    'age_category': category,
                    'health_needs': const.ALL,
                    'benefit':      const.ALL,
                    'status':       const.APPLICATION_STATUSES['IN_LINE_DEFERRED'],
                    'delivery':     const.ALL,
                    'group_type':   const.ALL,
                    'doo_type':     const.ALL,
                    'count':        queues['STATUS'][const.APPLICATION_STATUSES['IN_LINE_DEFERRED']][category]
                },
                #: Количество детей в очереди (с учетом детей, желающих сменить ДОО)
                {
                    'age_category': category,
                    'health_needs': const.ALL,
                    'benefit':      const.ALL,
                    'status':       const.APPLICATION_STATUSES['IN_LINE_WANT_CHANGE_DOO'],
                    'delivery':     const.ALL,
                    'group_type':   const.ALL,
                    'doo_type':     const.ALL,
                    'count':        queues['STATUS'][const.APPLICATION_STATUSES['IN_LINE_WANT_CHANGE_DOO']][category]
                },
                #: Количество детей в очереди с ограниченными возможностями здоровья
                {
                    'age_category': category,
                    'health_needs': const.HEALTH_NEEDS['DISABLED'],
                    'benefit':      const.ALL,
                    'status':       const.APPLICATION_STATUSES['IN_LINE'],
                    'delivery':     const.ALL,
                    'group_type':   const.ALL,
                    'doo_type':     const.ALL,
                    'count':        queues['HEALTH_NEEDS'][const.HEALTH_NEEDS['DISABLED']][category]
                },
                #: Количество детей в очереди в негосударственные ДОО
                {
                    'age_category': category,
                    'health_needs': const.ALL,
                    'benefit':      const.ALL,
                    'status':       const.APPLICATION_STATUSES['IN_LINE'],
                    'delivery':     const.ALL,
                    'group_type':   const.ALL,
                    'doo_type':     const.DOO_TYPES['NON_STATE'],
                    'count':        queues['DOO_TYPE'][const.DOO_TYPES['NON_STATE']][category]
                },
                #: Количество детей в очереди у ИП
                {
                    'age_category': category,
                    'health_needs': const.ALL,
                    'benefit':      const.ALL,
                    'status':       const.APPLICATION_STATUSES['IN_LINE'],
                    'delivery':     const.ALL,
                    'group_type':   const.ALL,
                    'doo_type':     const.DOO_TYPES['BUSINESS'],
                    'count':        queues['DOO_TYPE'][const.DOO_TYPES['BUSINESS']][category]
                },
                #: Количество детей в очереди, необеспеченных местом в ДОО
                {
                    'age_category': category,
                    'health_needs': const.ALL,
                    'benefit':      const.ALL,
                    'status':       const.APPLICATION_STATUSES['IN_LINE'],
                    'delivery':     const.ALL,
                    'group_type':   const.NO,
                    'doo_type':     const.NO,
                    'count':        queues['DOO_TYPE'][const.NO][category]
                },
                #: Количество детей в очереди, имеющих право на льготное зачисление
                {
                    'age_category': category,
                    'health_needs': const.ALL,
                    'benefit':      const.BENEFIT_TYPES['HAVE_BENEFITS'],
                    'status':       const.APPLICATION_STATUSES['IN_LINE'],
                    'delivery':     const.ALL,
                    'group_type':   const.ALL,
                    'doo_type':     const.ALL,
                    'count':        queues['BENEFITS'][const.BENEFIT_TYPES['HAVE_BENEFITS']][category]
                }
            ])

        #: Количество детей в очереди в группы кратковременного пребывания
        for category in const.AGE_CATEGORIES_CUT.values():
            result.append({
                'age_category': category,
                'health_needs': const.ALL,
                'benefit':      const.ALL,
                'status':       const.APPLICATION_STATUSES['IN_LINE'],
                'delivery':     const.ALL,
                'group_type':   const.GROUP_TYPES['SHORT_STAY'],
                'doo_type':     const.ALL,
                'count':        queues['GROUP_TYPE'][const.GROUP_TYPES['SHORT_STAY']][category]
            })
        for category in const.AGE_CATEGORIES_FULL_CURRENT.values():
            result.extend([
                # Отложенный спрос на текущую дату
                {
                    'age_category': const.AGE_CATEGORIES_FULL_MAP[category],
                    'health_needs': const.ALL,
                    'benefit':      const.ALL,
                    'doo_type':     const.ALL,
                    'group_type':   const.ALL,
                    'status':       const.APPLICATION_STATUSES['IN_LINE_DEFERRED'],
                    'delivery':     const.ALL,
                    'age_calculation_date': const.CURRENT_DATE_CALCULATION,
                    'count':        queues['STATUS'][const.APPLICATION_STATUSES['IN_LINE_DEFERRED']][category]
                },
                # Актуальный спрос на текущую дату
                {
                    'age_category': const.AGE_CATEGORIES_FULL_MAP[category],
                    'health_needs': const.ALL,
                    'benefit':      const.ALL,
                    'doo_type':     const.NO,
                    'group_type':   const.NO,
                    'status':       const.APPLICATION_STATUSES['IN_LINE'],
                    'delivery':     const.ALL,
                    'age_calculation_date': const.CURRENT_DATE_CALCULATION,
                    'count':        queues['DOO_TYPE'][const.NO][category]
                }
            ])
        result.extend([
            #: Количество детей в очереди, имеющих право на льготное зачисление на основании региональных льгот
            {
                'age_category': const.ALL,
                'health_needs': const.ALL,
                'benefit':      const.BENEFIT_TYPES['REGIONAL'],
                'status':       const.APPLICATION_STATUSES['IN_LINE'],
                'delivery':     const.ALL,
                'group_type':   const.ALL,
                'doo_type':     const.ALL,
                'count':        queues['BENEFITS'][const.BENEFIT_TYPES['REGIONAL']]
            },
            #: Количество детей в очереди, имеющих право на льготное зачисление на основании федеральных льгот
            {
                'age_category': const.ALL,
                'health_needs': const.ALL,
                'benefit':      const.BENEFIT_TYPES['FEDERAL'],
                'status':       const.APPLICATION_STATUSES['IN_LINE'],
                'delivery':     const.ALL,
                'group_type':   const.ALL,
                'doo_type':     const.ALL,
                'count':        queues['BENEFITS'][const.BENEFIT_TYPES['FEDERAL']]
            },
            #: Количество детей в очереди, нуждающихся в оздоровительных группах
            {
                'age_category': const.ALL,
                'health_needs': const.HEALTH_NEEDS['HEALTH'],
                'benefit':      const.ALL,
                'status':       const.APPLICATION_STATUSES['IN_LINE'],
                'delivery':     const.ALL,
                'group_type':   const.ALL,
                'doo_type':     const.ALL,
                'count':        queues['HEALTH_NEEDS'][const.HEALTH_NEEDS['HEALTH']]
            },
            #: Количество детей в очереди, нуждающихся в компенсирующих группах
            {
                'age_category': const.ALL,
                'health_needs': const.HEALTH_NEEDS['CORRECT'],
                'benefit':      const.ALL,
                'status':       const.APPLICATION_STATUSES['IN_LINE'],
                'delivery':     const.ALL,
                'group_type':   const.ALL,
                'doo_type':     const.ALL,
                'count':        queues['HEALTH_NEEDS'][const.HEALTH_NEEDS['CORRECT']]
            }
        ])
        return result

    @staticmethod
    def get_enrolled_data(enrolled):
        """
        Данные по зачисленным детям
        """
        result = []
        for category in const.AGE_CATEGORIES_FULL.values():
            result.extend([
                #: Количество зачисленных детей
                {
                    'age_category': category,
                    'health_needs': const.ALL,
                    'benefit':      const.ALL,
                    'status':       const.APPLICATION_STATUSES['ENROLLED'],
                    'delivery':     const.ALL,
                    'group_type':   const.ALL,
                    'doo_type':     const.ALL,
                    'count':        enrolled['COMMON'][category]
                },
                #: Количество детей, зачисленных в группы для детей с ограниченными возможностями здоровья
                {
                    'age_category': category,
                    'health_needs': const.ALL,
                    'benefit':      const.ALL,
                    'status':       const.APPLICATION_STATUSES['ENROLLED'],
                    'delivery':     const.ALL,
                    'group_type':   const.GROUP_TYPES['DISABLED'],
                    'doo_type':     const.ALL,
                    'count':        enrolled['GROUP_TYPE'][const.GROUP_TYPES['DISABLED']][category]
                },
                #: Количество детей, зачисленных в негосударственные ДОО
                {
                    'age_category': category,
                    'health_needs': const.ALL,
                    'benefit':      const.ALL,
                    'status':       const.APPLICATION_STATUSES['ENROLLED'],
                    'delivery':     const.ALL,
                    'group_type':   const.ALL,
                    'doo_type':     const.DOO_TYPES['NON_STATE'],
                    'count':        enrolled['DOO_TYPE'][const.DOO_TYPES['NON_STATE']][category]
                },
                #: Количество детей, зачисленных у ИП
                {
                    'age_category': category,
                    'health_needs': const.ALL,
                    'benefit':      const.ALL,
                    'status':       const.APPLICATION_STATUSES['ENROLLED'],
                    'delivery':     const.ALL,
                    'group_type':   const.ALL,
                    'doo_type':     const.DOO_TYPES['BUSINESS'],
                    'count':        enrolled['DOO_TYPE'][const.DOO_TYPES['BUSINESS']][category]
                },
                #: Количество детей
                {
                    'age_category': category,
                    'health_needs': const.ALL,
                    'benefit':      const.ALL,
                    'status':       const.APPLICATION_STATUSES['IN_LINE_ENROLLED'],
                    'delivery':     const.ALL,
                    'group_type':   const.ALL,
                    'doo_type':     const.ALL,
                    'count':        enrolled['STATUS'][const.APPLICATION_STATUSES['IN_LINE_ENROLLED']][category]
                }
            ])
        for category in const.AGE_CATEGORIES_CUT.values():
            result.extend([
                #: Количество детей, зачисленных в группы кратковременного пребывания
                {
                    'age_category': category,
                    'health_needs': const.ALL,
                    'benefit':      const.ALL,
                    'status':       const.APPLICATION_STATUSES['ENROLLED'],
                    'delivery':     const.ALL,
                    'group_type':   const.GROUP_TYPES['SHORT_STAY'],
                    'doo_type':     const.ALL,
                    'count':        enrolled['GROUP_TYPE'][const.GROUP_TYPES['SHORT_STAY']][category]
                },
                #: Количество детей, зачисленных в группы по присмотру и уходу в негосударственных ДОО
                {
                    'age_category': category,
                    'health_needs': const.ALL,
                    'benefit':      const.ALL,
                    'status':       const.APPLICATION_STATUSES['ENROLLED'],
                    'delivery':     const.ALL,
                    'group_type':   const.GROUP_TYPES['CARE'],
                    'doo_type':     const.DOO_TYPES['NON_STATE'],
                    'count':        enrolled['GROUP_TYPE'][const.GROUP_TYPES['CARE']][const.DOO_TYPES['NON_STATE']][category]
                },
                #: Количество детей, зачисленных в группы по присмотру и уходу в ИП
                {
                    'age_category': category,
                    'health_needs': const.ALL,
                    'benefit':      const.ALL,
                    'status':       const.APPLICATION_STATUSES['ENROLLED'],
                    'delivery':     const.ALL,
                    'group_type':   const.GROUP_TYPES['CARE'],
                    'doo_type':     const.DOO_TYPES['BUSINESS'],
                    'count':        enrolled['GROUP_TYPE'][const.GROUP_TYPES['CARE']][const.DOO_TYPES['BUSINESS']][category]
                }
            ])
        for category in const.AGE_CATEGORIES_FULL_CURRENT.values():
            result.append({
                'age_category':    const.AGE_CATEGORIES_FULL_MAP[category],
                'health_needs':     const.ALL,
                'benefit':         const.ALL,
                'doo_type':        const.ALL,
                'group_type':      const.ALL,
                'status':          const.APPLICATION_STATUSES['ENROLLED'],
                'delivery':        const.ALL,
                'age_calculation_date': const.CURRENT_DATE_CALCULATION,
                'count':           enrolled['COMMON'][category]
            })
        result.extend([
            #: Количество детей, желающие сменить ДОО
            {
                'age_category': const.ALL,
                'health_needs': const.ALL,
                'benefit':      const.ALL,
                'status':       const.APPLICATION_STATUSES['WANT_CHANGE_DOO'],
                'delivery':     const.ALL,
                'group_type':   const.ALL,
                'doo_type':     const.ALL,
                'count':        enrolled['STATUS'][const.APPLICATION_STATUSES['WANT_CHANGE_DOO']]
            },
            #: Количество детей, зачисленных в семейные дошкольные группы
            {
                'age_category': const.ALL,
                'health_needs': const.ALL,
                'benefit':      const.ALL,
                'status':       const.APPLICATION_STATUSES['ENROLLED'],
                'delivery':     const.ALL,
                'group_type':   const.GROUP_TYPES['FAMILY'],
                'doo_type':     const.ALL,
                'count':        enrolled['GROUP_TYPE'][const.GROUP_TYPES['FAMILY']]
            },
            #: Количество детей, зачисленных в компенсирующих группах
            {
                'age_category': const.ALL,
                'health_needs': const.ALL,
                'benefit':      const.ALL,
                'status':       const.APPLICATION_STATUSES['ENROLLED'],
                'delivery':     const.ALL,
                'group_type':   const.GROUP_TYPES['CORRECT'],
                'doo_type':     const.ALL,
                'count':        enrolled['GROUP_TYPE'][const.GROUP_TYPES['CORRECT']]
            },
            #: Количество детей, зачисленных в оздоровительных группах
            {
                'age_category': const.ALL,
                'health_needs': const.ALL,
                'benefit':      const.ALL,
                'status':       const.APPLICATION_STATUSES['ENROLLED'],
                'delivery':     const.ALL,
                'group_type':   const.GROUP_TYPES['HEALTH'],
                'doo_type':     const.ALL,
                'count':        enrolled['GROUP_TYPE'][const.GROUP_TYPES['HEALTH']]
            }
        ])
        return result

    @staticmethod
    def get_capacities_data(capacities):
        """
        Данные по свободным местам
        """
        result = []
        for category in const.AGE_CATEGORIES_CUT.values():
            result.extend([
                #: Количество свободных мест на дату сбора отчета
                {
                    'age_category': category,
                    'health_needs': const.ALL,
                    'benefit':      const.ALL,
                    'group_type':   const.ALL,
                    'doo_type':     const.ALL,
                    'count':        capacities['COMMON'][category]
                },
                #: Количество свободных мест в группах для детей с ограниченными возможностями здоровья
                {
                    'age_category': category,
                    'health_needs': const.ALL,
                    'benefit':      const.ALL,
                    'group_type':   const.GROUP_TYPES['DISABLED'],
                    'doo_type':     const.ALL,
                    'count':        capacities['GROUP_TYPE'][const.GROUP_TYPES['DISABLED']][category]
                },
                #: Количество свободных мест в негосударственных ДОО
                {
                    'age_category': category,
                    'health_needs': const.ALL,
                    'benefit':      const.ALL,
                    'group_type':   const.ALL,
                    'doo_type':     const.DOO_TYPES['NON_STATE'],
                    'count':        capacities['DOO_TYPE'][const.DOO_TYPES['NON_STATE']][category]
                },
                #: Количество свободных мест у ИП
                {
                    'age_category': category,
                    'health_needs': const.ALL,
                    'benefit':      const.ALL,
                    'group_type':   const.ALL,
                    'doo_type':     const.DOO_TYPES['BUSINESS'],
                    'count':        capacities['DOO_TYPE'][const.DOO_TYPES['BUSINESS']][category]
                },
                #: Количество свободных мест в группах кратковременного пребывания
                {
                    'age_category': category,
                    'health_needs': const.ALL,
                    'benefit':      const.ALL,
                    'group_type':   const.GROUP_TYPES['SHORT_STAY'],
                    'doo_type':     const.ALL,
                    'count':        capacities['GROUP_TYPE'][const.GROUP_TYPES['SHORT_STAY']][category]
                }])
        return result

    def collect_data(self, unit):
        """
        Подсчет показателей по учреждению
        """
        applications = queues = enrolled = capacities = None
        report_form_row, created = ReportFormRow.objects.get_or_create(report=self.report_form, unit=unit)
        if unit.kind.id == UnitKind.DOU:
            applications = self.collect_applications_data(unit)
            queues = self.collect_queues_data(unit)
            enrolled = self.collect_enrolled_data(unit)
            capacities = self.collect_capacities_data(unit)

        elif unit.kind.id == UnitKind.MO:
            applications = self.collect_applications_data(unit) if gs_settings.NO_CONSIDER_PRIORITY else self.cache_mo.applications
            queues = self.collect_queues_data(unit) if gs_settings.NO_CONSIDER_PRIORITY else self.cache_mo.queues
            enrolled = self.cache_mo.enrolled
            # TODO: Большой костыль (#84427)
            for key, value in const.AGE_CATEGORIES_FULL.iteritems():
                enrolled['STATUS'][const.APPLICATION_STATUSES['IN_LINE_ENROLLED']][value] = (enrolled['COMMON'][value] +
                                                                                queues['DOO_TYPE'][const.NO][value])
            capacities = self.cache_mo.capacities

            if self.is_region:
                deepupdate(applications, self.cache_region.applications)
                deepupdate(queues, self.cache_region.queues)
                deepupdate(enrolled, self.cache_region.enrolled)
                deepupdate(capacities, self.cache_region.capacities)

            self.cache_mo = row_cache({}, {}, {}, {})

        elif unit.kind.id == UnitKind.REGION:
            applications = self.cache_region.applications
            queues = self.cache_region.queues
            enrolled = self.cache_region.enrolled
            capacities = self.cache_region.capacities

        report_form_row.applications = json.dumps(applications)
        report_form_row.queues = json.dumps(queues)
        report_form_row.enrolled = json.dumps(enrolled)
        report_form_row.capacities = json.dumps(capacities)
        report_form_row.save()

    def get_param(self, _id, func, prefix):
        """
        получения значения из кеша
        @param _id: int идетификатор котрый ищем в кеше
        @param func: callable Объект который должен вернуть значения для кеша
        @param prefix: str имя кеша
        """
        if not prefix in self.cache:
            self.cache[prefix] = {}
        if not _id in self.cache[prefix]:
            self.cache[prefix][_id] = func(_id)
        return self.cache[prefix][_id]

    @staticmethod
    def _create_loc(unit, dct):
        """
        Создание заголовка locality
        """
        kind_id = unit.kind.id
        if kind_id == UnitKind.DOU:
            try:
                code = unit.gisdo.doo_identity
            except GisdoUnit.DoesNotExist:
                raise ApplicationLogicException(u'Учреждение c ID=%d не найдено'
                                                u' в модуле "Фед. отчетности"'
                                                % unit.pk)
            dct['unit'] = {'code': code, 'name': unit.name}
            dct.setdefault('district', {'code': ''})
            dct.setdefault('settlement', {'code': ''})
        elif kind_id == UnitKind.RAION:
            dct['district'] = {'code': unit.code or unit.name, 'name': unit.name}
            dct.setdefault('settlement', {'code': ''})
        elif kind_id == UnitKind.PLACE:
            dct['settlement'] = {'code': unit.code or unit.name, 'name': unit.name}
        elif kind_id == UnitKind.MO:
            dct['municipality'] = {
                'code':     unit.code or unit.name,
                'name':     unit.name or '',
                'oktmo':    getattr(unit, 'octmo', '')
            }
        elif kind_id == UnitKind.REGION:
            dct['state'] = {'code': unit.code or unit.name, 'name': unit.name}

        if unit.parent:
            # Например для филиала.
            if unit.parent.kind.id == UnitKind.DOU:
                ReportFormHelper._create_loc(unit.get_mo(), dct)
            else:
                ReportFormHelper._create_loc(unit.parent, dct)

    @staticmethod
    def create_locality(unit):
        """
        Создание данных для locality
        """
        locality = {
            'federal_district': {
                'code': const.FEDERAL_DISTRICTS[gs_settings.FEDERAL_DISTRICT],
                'name': const.FEDERAL_DISTRICTS[gs_settings.FEDERAL_DISTRICT]
            },
            'state': {'code': ''},
            'section': {'code': ''},
            'municipality': {'code': ''}
        }
        ReportFormHelper._create_loc(unit, locality)
        return locality

    def collect_reports(self, mo):
        """
        Подсчет показателей по дочерним учреждениям МО
        """
        for unit in ReportFormHelper.get_list_dou(mo):
            yield unit, self.collect_data(unit)

    def get_reports(self, mo):
        """
        Создание отчетов по дочерним учреждениям МО
        """
        for unit in ReportFormHelper.get_list_dou(mo):
            yield unit, self.create_sub_report(unit)

    def create_sub_report(self, unit):
        """
        подсчет показателей, сохранение и формирование для шаблона
        @param unit ссылка на Unit, учреждение по котрому считаем отчет
        """
        try:
            report_form_row = ReportFormRow.objects.get(report=self.report_form, unit=unit)
        except ReportFormRow.DoesNotExist:
            log.error(traceback.format_exc())
            return None
        applications_data = self.get_applications_data(json.loads(report_form_row.applications))
        queues_data = self.get_queues_data(json.loads(report_form_row.queues))
        enrolled_data = self.get_enrolled_data(json.loads(report_form_row.enrolled))
        capacities_data = self.get_capacities_data(json.loads(report_form_row.capacities))

        result = {
            'assertions': applications_data + queues_data + enrolled_data,
            'capacities': capacities_data
        }

        result.update(self.create_locality(unit))
        return result

    def create_client(self):
        """
        Создание клиента веб-сервиса
        """
        try:
            header = Element('ns1:Authentication')
            header.append(Element('ns1:Login').setText(self.schedule_settings.push_login))
            header.append(Element('ns1:Password').setText(self.schedule_settings.push_password))
            client = Client(url=gs_settings.GISDO_WS_WSDL, timeout=gs_settings.GISDO_TIMEOUT)
            client.set_options(soapheaders=(header,))
        except Exception, error:
            raise SendReportException(error.message)
        return client

    def render_xml(self, report_data):
        """
        Рендер xml
        """
        xml_string = loader.render_to_string(self.template, report_data).encode('utf-8')
        if self.schedule_settings.zip_xml:
            xml_string = zlib.compress(xml_string)[2:-4]
        encoded_xml = base64.b64encode(xml_string)
        return encoded_xml

    def send(self, client, encoded_xml):
        """
        Отправка отчетов на веб-сервис
        """
        error_msg = None
        if self.schedule_settings.zip_xml:
            if self.schedule_settings.async_send:
                method = 'PushZipDataAsync'
            else:
                method = 'PushZipData'
        else:
            method = 'PushData'

        sent, sent_iterator = False, 1

        while not sent and sent_iterator < int(self.schedule_settings.resend_count):
            try:
                result = getattr(client.service, method)(data=encoded_xml, schema=u'default')
            except (socket.timeout, suds.WebFault) as error:
                error_msg = str(error).encode('utf-8')
                sent_iterator += 1
            except Exception, error:
                error_msg = str(error).encode('utf-8')
                break
            else:
                if not result['Result']:
                    error_msg = result['Message']
                    break
                else:
                    sent = True

        if error_msg:
            raise SendReportException(error_msg)